from importlib.metadata import version as v
from pathlib import PurePath
from re import compile as regex
from sys import path

path.insert(0, str(PurePath(__file__).parents[2] / "tree_sitter"))

project = "py-tree-sitter"
author = "Max Brunsfeld"
copyright = "2019, MIT license"
release = v("tree_sitter")

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
]
source_suffix = {
    ".rst": "restructuredtext"
}
master_doc = "index"
language = "en"
needs_sphinx = "8.1"
templates_path = ["_templates"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.10/", None),
}

autoclass_content = "class"
autodoc_member_order = "alphabetical"
autosummary_generate = False

napoleon_numpy_docstring = True
napoleon_google_docstring = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_use_admonition_for_notes = True

html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/tree-sitter/py-tree-sitter",
    "pygments_light_style": "default",
    "pygments_dark_style": "github-dark",
    "navigation_with_keys": False,
    "use_repository_button": True,
    "use_download_button": False,
    "use_fullscreen_button": False,
    "show_toc_level": 2,
}
html_static_path = ["_static"]
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"


special_doc = regex(r"\S*self[^.]+")


def process_signature(_app, _what, name, _obj, _options, _signature, return_annotation):
    if name == "tree_sitter.Language":
        return "(ptr)", return_annotation
    if name == "tree_sitter.Query":
        return "(language, source)", return_annotation
    if name == "tree_sitter.QueryCursor":
        return "(query, *, match_limit=None, timeout_micros=None)", return_annotation
    if name == "tree_sitter.Parser":
        return "(language, *, included_ranges=None, timeout_micros=None)", return_annotation
    if name == "tree_sitter.Range":
        return "(start_point, end_point, start_byte, end_byte)", return_annotation
    if name == "tree_sitter.QueryPredicate":
        return None, return_annotation
    if name == "tree_sitter.LogType":
        return None, return_annotation


def process_docstring(_app, what, name, _obj, _options, lines):
    if what == "data":
        lines.clear()
    elif what == "method":
        if name.endswith("__index__"):
            lines[0] = "Converts ``self`` to an integer for use as an index."
        elif name.endswith("__") and lines and "self" in lines[0]:
            lines[0] = f"Implements ``{special_doc.search(lines[0]).group(0)}``."


def process_bases(_app, name, _obj, _options, bases):
    if name == "tree_sitter.Point":
        bases[-1] = ":class:`~typing.NamedTuple`"
    if name == "tree_sitter.LogType":
        bases[-1] = ":class:`~enum.IntEnum`"
    if name == "tree_sitter.LookaheadIterator":
        bases[-1] = ":class:`~collections.abc.Iterator`"


def setup(app):
    app.connect("autodoc-process-signature", process_signature)
    app.connect("autodoc-process-docstring", process_docstring)
    app.connect("autodoc-process-bases", process_bases)



from sys import maxsize
from typing import cast
from unittest import TestCase

from tree_sitter import Language

import tree_sitter_html
import tree_sitter_javascript
import tree_sitter_json
import tree_sitter_python
import tree_sitter_rust


class TestLanguage(TestCase):
    def setUp(self):
        self.html = tree_sitter_html.language()
        self.javascript = tree_sitter_javascript.language()
        self.json = tree_sitter_json.language()
        self.python = tree_sitter_python.language()
        self.rust = tree_sitter_rust.language()

    def test_init_invalid(self):
        self.assertRaises(ValueError, Language, 42)

    def test_properties(self):
        lang = Language(self.python)
        self.assertEqual(lang.abi_version, 14)
        self.assertEqual(lang.node_kind_count, 275)
        self.assertEqual(lang.parse_state_count, 2809)
        self.assertEqual(lang.field_count, 32)

    def test_node_kind_for_id(self):
        lang = Language(self.json)
        self.assertEqual(lang.node_kind_for_id(1), "{")
        self.assertEqual(lang.node_kind_for_id(3), "}")

    def test_id_for_node_kind(self):
        lang = Language(self.json)
        self.assertEqual(lang.id_for_node_kind(":", False), 4)
        self.assertEqual(lang.id_for_node_kind("string", True), 20)

    def test_node_kind_is_named(self):
        lang = Language(self.json)
        self.assertFalse(lang.node_kind_is_named(4))
        self.assertTrue(lang.node_kind_is_named(20))

    def test_node_kind_is_visible(self):
        lang = Language(self.json)
        self.assertTrue(lang.node_kind_is_visible(2))

    def test_field_name_for_id(self):
        lang = Language(self.json)
        self.assertEqual(lang.field_name_for_id(1), "key")
        self.assertEqual(lang.field_name_for_id(2), "value")

    def test_field_id_for_name(self):
        lang = Language(self.json)
        self.assertEqual(lang.field_id_for_name("key"), 1)
        self.assertEqual(lang.field_id_for_name("value"), 2)

    def test_next_state(self):
        lang = Language(self.javascript)
        self.assertNotEqual(lang.next_state(1, 1), 0)

    def test_lookahead_iterator(self):
        lang = Language(self.javascript)
        self.assertIsNotNone(lang.lookahead_iterator(0))
        self.assertIsNone(lang.lookahead_iterator(9999))

    def test_eq(self):
        self.assertEqual(Language(self.json), Language(self.json))
        self.assertNotEqual(Language(self.rust), Language(self.html))

    def test_hash(self):
        for name in ["html", "javascript", "json", "python", "rust"]:
            with self.subTest(language=name):
                lang = Language(cast(object, getattr(self, name)))
                self.assertGreater(hash(lang) & maxsize << 1, 0)



from typing import cast
from unittest import TestCase

import tree_sitter_python
import tree_sitter_javascript
import tree_sitter_json

from tree_sitter import Language, Node, Parser

JSON_EXAMPLE = b"""

[
  123,
  false,
  {
    "x": null
  }
]
"""


def get_all_nodes(tree):
    result = []
    visited_children = False
    cursor = tree.walk()
    while True:
        if not visited_children:
            result.append(cursor.node)
            if not cursor.goto_first_child():
                visited_children = True
        elif cursor.goto_next_sibling():
            visited_children = False
        elif not cursor.goto_parent():
            break
    return result


class TestNode(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.javascript = Language(tree_sitter_javascript.language())
        cls.json = Language(tree_sitter_json.language())
        cls.python = Language(tree_sitter_python.language())

    def test_child_by_field_id(self):
        parser = Parser(self.python)
        tree = parser.parse(b"def foo():\n  bar()")
        root_node = tree.root_node
        fn_node = tree.root_node.children[0]

        self.assertIsNone(self.python.field_id_for_name("noname"))
        name_field = cast(int, self.python.field_id_for_name("name"))
        alias_field = cast(int, self.python.field_id_for_name("alias"))
        self.assertIsNone(root_node.child_by_field_id(alias_field))
        self.assertIsNone(root_node.child_by_field_id(name_field))
        self.assertIsNone(fn_node.child_by_field_id(alias_field))
        self.assertIsNone(fn_node.child_by_field_name("noname"))
        self.assertEqual(fn_node.child_by_field_name("name"), fn_node.child_by_field_name("name"))

    def test_child_by_field_name(self):
        parser = Parser(self.python)
        tree = parser.parse(b"while a:\n  pass")
        while_node = cast(Node, tree.root_node.child(0))
        self.assertIsNotNone(while_node)
        self.assertEqual(while_node.type, "while_statement")
        self.assertEqual(while_node.child_by_field_name("body"), while_node.child(3))

    def test_children_by_field_id(self):
        parser = Parser(self.javascript)
        tree = parser.parse(b"<div a={1} b={2} />")
        jsx_node = tree.root_node.children[0].children[0]
        attribute_field = cast(int, self.javascript.field_id_for_name("attribute"))
        attributes = jsx_node.children_by_field_id(attribute_field)
        self.assertListEqual([a.type for a in attributes], ["jsx_attribute", "jsx_attribute"])

    def test_children_by_field_name(self):
        parser = Parser(self.javascript)
        tree = parser.parse(b"<div a={1} b={2} />")
        jsx_node = tree.root_node.children[0].children[0]
        attributes = jsx_node.children_by_field_name("attribute")
        self.assertListEqual([a.type for a in attributes], ["jsx_attribute", "jsx_attribute"])

    def test_field_name_for_child(self):
        parser = Parser(self.javascript)
        tree = parser.parse(b"<div a={1} b={2} />")
        jsx_node = tree.root_node.children[0].children[0]

        self.assertIsNone(jsx_node.field_name_for_child(0))
        self.assertEqual(jsx_node.field_name_for_child(1), "name")

    def test_root_node_with_offset(self):
        parser = Parser(self.javascript)
        tree = parser.parse(b"  if (a) b")

        node = cast(Node, tree.root_node_with_offset(6, (2, 2)))
        self.assertIsNotNone(node)
        self.assertEqual(node.byte_range, (8, 16))
        self.assertEqual(node.start_point, (2, 4))
        self.assertEqual(node.end_point, (2, 12))

        child = cast(Node, cast(Node, node.child(0)).child(2))
        self.assertIsNotNone(child)
        self.assertEqual(child.type, "expression_statement")
        self.assertEqual(child.byte_range, (15, 16))
        self.assertEqual(child.start_point, (2, 11))
        self.assertEqual(child.end_point, (2, 12))

        cursor = node.walk()
        cursor.goto_first_child()
        cursor.goto_first_child()
        cursor.goto_next_sibling()
        child = cast(Node, cursor.node)
        self.assertIsNotNone(child)
        self.assertEqual(child.type, "parenthesized_expression")
        self.assertEqual(child.byte_range, (11, 14))
        self.assertEqual(child.start_point, (2, 7))
        self.assertEqual(child.end_point, (2, 10))

    def test_descendant_count(self):
        parser = Parser(self.json)
        tree = parser.parse(JSON_EXAMPLE)
        value_node = tree.root_node
        all_nodes = get_all_nodes(tree)

        self.assertEqual(value_node.descendant_count, len(all_nodes))

        cursor = value_node.walk()
        for i, node in enumerate(all_nodes):
            cursor.goto_descendant(i)
            self.assertEqual(cursor.node, node, f"index {i}")

        for i, node in reversed(list(enumerate(all_nodes))):
            cursor.goto_descendant(i)
            self.assertEqual(cursor.node, node, f"rev index {i}")

    def test_descendant_for_byte_range(self):
        parser = Parser(self.json)
        tree = parser.parse(JSON_EXAMPLE)
        array_node = tree.root_node

        colon_index = JSON_EXAMPLE.index(b":")

        # Leaf node exactly matches the given bounds - byte query
        colon_node = array_node.descendant_for_byte_range(colon_index, colon_index + 1)
        self.assertIsNotNone(colon_node)
        colon_node = cast(Node, colon_node)
        self.assertEqual(colon_node.type, ":")
        self.assertEqual(colon_node.start_byte, colon_index)
        self.assertEqual(colon_node.end_byte, colon_index + 1)
        self.assertEqual(colon_node.start_point, (6, 7))
        self.assertEqual(colon_node.end_point, (6, 8))

        # Leaf node exactly matches the given bounds - point query
        colon_node = array_node.descendant_for_point_range((6, 7), (6, 8))
        self.assertIsNotNone(colon_node)
        colon_node = cast(Node, colon_node)
        self.assertEqual(colon_node.type, ":")
        self.assertEqual(colon_node.start_byte, colon_index)
        self.assertEqual(colon_node.end_byte, colon_index + 1)
        self.assertEqual(colon_node.start_point, (6, 7))
        self.assertEqual(colon_node.end_point, (6, 8))

        # The given point is between two adjacent leaf nodes - byte query
        colon_node = array_node.descendant_for_byte_range(colon_index, colon_index)
        self.assertIsNotNone(colon_node)
        colon_node = cast(Node, colon_node)
        self.assertEqual(colon_node.type, ":")
        self.assertEqual(colon_node.start_byte, colon_index)
        self.assertEqual(colon_node.end_byte, colon_index + 1)
        self.assertEqual(colon_node.start_point, (6, 7))
        self.assertEqual(colon_node.end_point, (6, 8))

        # The given point is between two adjacent leaf nodes - point query
        colon_node = array_node.descendant_for_point_range((6, 7), (6, 7))
        self.assertIsNotNone(colon_node)
        colon_node = cast(Node, colon_node)
        self.assertEqual(colon_node.type, ":")
        self.assertEqual(colon_node.start_byte, colon_index)
        self.assertEqual(colon_node.end_byte, colon_index + 1)
        self.assertEqual(colon_node.start_point, (6, 7))
        self.assertEqual(colon_node.end_point, (6, 8))

        # Leaf node starts at the lower bound, ends after the upper bound - byte query
        string_index = JSON_EXAMPLE.index(b'"x"')
        string_node = array_node.descendant_for_byte_range(string_index, string_index + 2)
        self.assertIsNotNone(string_node)
        string_node = cast(Node, string_node)
        self.assertEqual(string_node.type, "string")
        self.assertEqual(string_node.start_byte, string_index)
        self.assertEqual(string_node.end_byte, string_index + 3)
        self.assertEqual(string_node.start_point, (6, 4))
        self.assertEqual(string_node.end_point, (6, 7))

        # Leaf node starts at the lower bound, ends after the upper bound - point query
        string_node = array_node.descendant_for_point_range((6, 4), (6, 6))
        self.assertIsNotNone(string_node)
        string_node = cast(Node, string_node)
        self.assertEqual(string_node.type, "string")
        self.assertEqual(string_node.start_byte, string_index)
        self.assertEqual(string_node.end_byte, string_index + 3)
        self.assertEqual(string_node.start_point, (6, 4))
        self.assertEqual(string_node.end_point, (6, 7))

        # Leaf node starts before the lower bound, ends at the upper bound - byte query
        null_index = JSON_EXAMPLE.index(b"null")
        null_node = array_node.descendant_for_byte_range(null_index + 1, null_index + 4)
        self.assertIsNotNone(null_node)
        null_node = cast(Node, null_node)
        self.assertEqual(null_node.type, "null")
        self.assertEqual(null_node.start_byte, null_index)
        self.assertEqual(null_node.end_byte, null_index + 4)
        self.assertEqual(null_node.start_point, (6, 9))
        self.assertEqual(null_node.end_point, (6, 13))

        # Leaf node starts before the lower bound, ends at the upper bound - point query
        null_node = array_node.descendant_for_point_range((6, 11), (6, 13))
        self.assertIsNotNone(null_node)
        null_node = cast(Node, null_node)
        self.assertEqual(null_node.type, "null")
        self.assertEqual(null_node.start_byte, null_index)
        self.assertEqual(null_node.end_byte, null_index + 4)
        self.assertEqual(null_node.start_point, (6, 9))
        self.assertEqual(null_node.end_point, (6, 13))

        # The bounds span multiple leaf nodes - return the smallest node that does span it.
        pair_node = array_node.descendant_for_byte_range(string_index + 2, string_index + 4)
        self.assertIsNotNone(pair_node)
        pair_node = cast(Node, pair_node)
        self.assertEqual(pair_node.type, "pair")
        self.assertEqual(pair_node.start_byte, string_index)
        self.assertEqual(pair_node.end_byte, string_index + 9)
        self.assertEqual(pair_node.start_point, (6, 4))
        self.assertEqual(pair_node.end_point, (6, 13))

        self.assertEqual(colon_node.parent, pair_node)

        # No leaf spans the given range - return the smallest node that does span it.
        pair_node = array_node.descendant_for_point_range((6, 6), (6, 8))
        self.assertIsNotNone(pair_node)
        pair_node = cast(Node, pair_node)
        self.assertEqual(pair_node.type, "pair")
        self.assertEqual(pair_node.start_byte, string_index)
        self.assertEqual(pair_node.end_byte, string_index + 9)
        self.assertEqual(pair_node.start_point, (6, 4))
        self.assertEqual(pair_node.end_point, (6, 13))

    def test_children(self):
        parser = Parser(self.python)
        tree = parser.parse(b"def foo():\n  bar()")

        root_node = tree.root_node
        self.assertEqual(root_node.type, "module")
        self.assertEqual(root_node.start_byte, 0)
        self.assertEqual(root_node.end_byte, 18)
        self.assertEqual(root_node.start_point, (0, 0))
        self.assertEqual(root_node.end_point, (1, 7))

        # List object is reused
        self.assertIs(root_node.children, root_node.children)

        fn_node = root_node.children[0]
        self.assertEqual(fn_node, root_node.child(0))
        self.assertEqual(fn_node.type, "function_definition")
        self.assertEqual(fn_node.start_byte, 0)
        self.assertEqual(fn_node.end_byte, 18)
        self.assertEqual(fn_node.start_point, (0, 0))
        self.assertEqual(fn_node.end_point, (1, 7))

        def_node = fn_node.children[0]
        self.assertEqual(def_node, fn_node.child(0))
        self.assertEqual(def_node.type, "def")
        self.assertEqual(def_node.is_named, False)

        id_node = fn_node.children[1]
        self.assertEqual(id_node, fn_node.child(1))
        self.assertEqual(id_node.type, "identifier")
        self.assertEqual(id_node.is_named, True)
        self.assertEqual(len(id_node.children), 0)

        params_node = fn_node.children[2]
        self.assertEqual(params_node, fn_node.child(2))
        self.assertEqual(params_node.type, "parameters")
        self.assertEqual(params_node.is_named, True)

        colon_node = fn_node.children[3]
        self.assertEqual(colon_node, fn_node.child(3))
        self.assertEqual(colon_node.type, ":")
        self.assertEqual(colon_node.is_named, False)

        statement_node = fn_node.children[4]
        self.assertEqual(statement_node, fn_node.child(4))
        self.assertEqual(statement_node.type, "block")
        self.assertEqual(statement_node.is_named, True)

    def test_is_extra(self):
        parser = Parser(self.javascript)
        tree = parser.parse(b"foo(/* hi */);")

        root_node = tree.root_node
        comment_node = cast(Node, root_node.descendant_for_byte_range(7, 7))
        self.assertIsNotNone(comment_node)

        self.assertEqual(root_node.type, "program")
        self.assertEqual(comment_node.type, "comment")
        self.assertEqual(root_node.is_extra, False)
        self.assertEqual(comment_node.is_extra, True)

    def test_properties(self):
        parser = Parser(self.python)
        tree = parser.parse(b"[1, 2, 3]")

        root_node = tree.root_node
        self.assertEqual(root_node.type, "module")
        self.assertEqual(root_node.start_byte, 0)
        self.assertEqual(root_node.end_byte, 9)
        self.assertEqual(root_node.start_point, (0, 0))
        self.assertEqual(root_node.end_point, (0, 9))

        exp_stmt_node = root_node.children[0]
        self.assertEqual(exp_stmt_node, root_node.child(0))
        self.assertEqual(exp_stmt_node.type, "expression_statement")
        self.assertEqual(exp_stmt_node.start_byte, 0)
        self.assertEqual(exp_stmt_node.end_byte, 9)
        self.assertEqual(exp_stmt_node.start_point, (0, 0))
        self.assertEqual(exp_stmt_node.end_point, (0, 9))
        self.assertEqual(exp_stmt_node.parent, root_node)

        list_node = exp_stmt_node.children[0]
        self.assertEqual(list_node, exp_stmt_node.child(0))
        self.assertEqual(list_node.type, "list")
        self.assertEqual(list_node.start_byte, 0)
        self.assertEqual(list_node.end_byte, 9)
        self.assertEqual(list_node.start_point, (0, 0))
        self.assertEqual(list_node.end_point, (0, 9))
        self.assertEqual(list_node.parent, exp_stmt_node)

        named_children = list_node.named_children

        open_delim_node = list_node.children[0]
        self.assertEqual(open_delim_node, list_node.child(0))
        self.assertEqual(open_delim_node.type, "[")
        self.assertEqual(open_delim_node.start_byte, 0)
        self.assertEqual(open_delim_node.end_byte, 1)
        self.assertEqual(open_delim_node.start_point, (0, 0))
        self.assertEqual(open_delim_node.end_point, (0, 1))
        self.assertEqual(open_delim_node.parent, list_node)

        first_num_node = list_node.children[1]
        self.assertEqual(first_num_node, list_node.child(1))
        self.assertEqual(first_num_node, open_delim_node.next_named_sibling)
        self.assertEqual(first_num_node.parent, list_node)
        self.assertEqual(named_children[0], first_num_node)
        self.assertEqual(first_num_node, list_node.named_child(0))

        first_comma_node = list_node.children[2]
        self.assertEqual(first_comma_node, list_node.child(2))
        self.assertEqual(first_comma_node, first_num_node.next_sibling)
        self.assertEqual(first_num_node, first_comma_node.prev_sibling)
        self.assertEqual(first_comma_node.parent, list_node)

        second_num_node = list_node.children[3]
        self.assertEqual(second_num_node, list_node.child(3))
        self.assertEqual(second_num_node, first_comma_node.next_sibling)
        self.assertEqual(second_num_node, first_num_node.next_named_sibling)
        self.assertEqual(first_num_node, second_num_node.prev_named_sibling)
        self.assertEqual(second_num_node.parent, list_node)
        self.assertEqual(named_children[1], second_num_node)
        self.assertEqual(second_num_node, list_node.named_child(1))

        second_comma_node = list_node.children[4]
        self.assertEqual(second_comma_node, list_node.child(4))
        self.assertEqual(second_comma_node, second_num_node.next_sibling)
        self.assertEqual(second_num_node, second_comma_node.prev_sibling)
        self.assertEqual(second_comma_node.parent, list_node)

        third_num_node = list_node.children[5]
        self.assertEqual(third_num_node, list_node.child(5))
        self.assertEqual(third_num_node, second_comma_node.next_sibling)
        self.assertEqual(third_num_node, second_num_node.next_named_sibling)
        self.assertEqual(second_num_node, third_num_node.prev_named_sibling)
        self.assertEqual(third_num_node.parent, list_node)
        self.assertEqual(named_children[2], third_num_node)
        self.assertEqual(third_num_node, list_node.named_child(2))

        close_delim_node = list_node.children[6]
        self.assertEqual(close_delim_node, list_node.child(6))
        self.assertEqual(close_delim_node.type, "]")
        self.assertEqual(close_delim_node.start_byte, 8)
        self.assertEqual(close_delim_node.end_byte, 9)
        self.assertEqual(close_delim_node.start_point, (0, 8))
        self.assertEqual(close_delim_node.end_point, (0, 9))
        self.assertEqual(close_delim_node, third_num_node.next_sibling)
        self.assertEqual(third_num_node, close_delim_node.prev_sibling)
        self.assertEqual(third_num_node, close_delim_node.prev_named_sibling)
        self.assertEqual(close_delim_node.parent, list_node)

        self.assertEqual(list_node.child_count, 7)
        self.assertEqual(list_node.named_child_count, 3)

    def test_numeric_symbols_respect_simple_aliases(self):
        parser = Parser(self.python)

        # Example 1:
        # Python argument lists can contain "splat" arguments, which are not allowed within
        # other expressions. This includes `parenthesized_list_splat` nodes like `(*b)`. These
        # `parenthesized_list_splat` nodes are aliased as `parenthesized_expression`. Their numeric
        # `symbol`, aka `kind_id` should match that of a normal `parenthesized_expression`.
        tree = parser.parse(b"(a((*b)))")
        root_node = tree.root_node
        self.assertEqual(
            str(root_node),
            "(module (expression_statement (parenthesized_expression (call "
            + "function: (identifier) arguments: (argument_list (parenthesized_expression "
            + "(list_splat (identifier))))))))",
        )

        outer_expr_node = cast(Node, root_node.child(0)).child(0)
        self.assertIsNotNone(outer_expr_node)
        outer_expr_node = cast(Node, outer_expr_node)
        self.assertEqual(outer_expr_node.type, "parenthesized_expression")

        inner_expr_node = cast(Node, outer_expr_node.named_child(0))
        inner_expr_node = cast(Node, inner_expr_node.child_by_field_name("arguments"))
        inner_expr_node = cast(Node, inner_expr_node.named_child(0))
        self.assertIsNotNone(inner_expr_node)

        self.assertEqual(inner_expr_node.type, "parenthesized_expression")
        self.assertEqual(inner_expr_node.kind_id, outer_expr_node.kind_id)

    def test_tree(self):
        code = b"def foo():\n  bar()\n\ndef foo():\n  bar()"
        parser = Parser(self.python)

        for item in parser.parse(code).root_node.children:
            self.assertIsNotNone(item.is_named)

        for item in parser.parse(code).root_node.children:
            self.assertIsNotNone(item.is_named)

    def test_text(self):
        parser = Parser(self.python)
        tree = parser.parse(b"[0, [1, 2, 3]]")

        root_node = tree.root_node
        self.assertEqual(root_node.text, b"[0, [1, 2, 3]]")

        exp_stmt_node = root_node.children[0]
        self.assertEqual(exp_stmt_node.text, b"[0, [1, 2, 3]]")

        list_node = exp_stmt_node.children[0]
        self.assertEqual(list_node.text, b"[0, [1, 2, 3]]")

        open_delim_node = list_node.children[0]
        self.assertEqual(open_delim_node.text, b"[")

        first_num_node = list_node.children[1]
        self.assertEqual(first_num_node.text, b"0")

        first_comma_node = list_node.children[2]
        self.assertEqual(first_comma_node.text, b",")

        child_list_node = list_node.children[3]
        self.assertEqual(child_list_node.text, b"[1, 2, 3]")

        close_delim_node = list_node.children[4]
        self.assertEqual(close_delim_node.text, b"]")

    def test_hash(self):
        parser = Parser(self.python)
        source_code = b"def foo():\n  bar()\n  bar()"
        tree = parser.parse(source_code)
        root_node = tree.root_node
        first_function_node = root_node.children[0]
        second_function_node = root_node.children[0]

        # Uniqueness and consistency
        self.assertEqual(hash(first_function_node), hash(first_function_node))
        self.assertNotEqual(hash(root_node), hash(first_function_node))

        # Equality implication
        self.assertEqual(hash(first_function_node), hash(second_function_node))
        self.assertEqual(first_function_node, second_function_node)

        # Different nodes with different properties
        different_tree = parser.parse(b"def baz():\n  qux()")
        different_node = different_tree.root_node.children[0]
        self.assertNotEqual(hash(first_function_node), hash(different_node))

        # Same code, different parse trees
        another_tree = parser.parse(source_code)
        another_node = another_tree.root_node.children[0]
        self.assertNotEqual(hash(first_function_node), hash(another_node))



from typing import cast
from unittest import TestCase

from tree_sitter import Language, Node, Parser

import tree_sitter_python
import tree_sitter_rust


class TestTree(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.python = Language(tree_sitter_python.language())
        cls.rust = Language(tree_sitter_rust.language())

    def test_edit(self):
        parser = Parser(self.python)
        tree = parser.parse(b"def foo():\n  bar()")

        edit_offset = len(b"def foo(")
        tree.edit(
            start_byte=edit_offset,
            old_end_byte=edit_offset,
            new_end_byte=edit_offset + 2,
            start_point=(0, edit_offset),
            old_end_point=(0, edit_offset),
            new_end_point=(0, edit_offset + 2),
        )

        fn_node = tree.root_node.children[0]
        self.assertEqual(fn_node.type, "function_definition")
        self.assertTrue(fn_node.has_changes)
        self.assertFalse(fn_node.children[0].has_changes)
        self.assertFalse(fn_node.children[1].has_changes)
        self.assertFalse(fn_node.children[3].has_changes)

        params_node = fn_node.children[2]
        self.assertEqual(params_node.type, "parameters")
        self.assertTrue(params_node.has_changes)
        self.assertEqual(params_node.start_point, (0, edit_offset - 1))
        self.assertEqual(params_node.end_point, (0, edit_offset + 3))

        new_tree = parser.parse(b"def foo(ab):\n  bar()", tree)
        self.assertEqual(
            str(new_tree.root_node),
            "(module (function_definition"
            + " name: (identifier)"
            + " parameters: (parameters (identifier))"
            + " body: (block"
            + " (expression_statement (call"
            + " function: (identifier)"
            + " arguments: (argument_list))))))",
        )

    def test_changed_ranges(self):
        parser = Parser(self.python)
        tree = parser.parse(b"def foo():\n  bar()")

        edit_offset = len(b"def foo(")
        tree.edit(
            start_byte=edit_offset,
            old_end_byte=edit_offset,
            new_end_byte=edit_offset + 2,
            start_point=(0, edit_offset),
            old_end_point=(0, edit_offset),
            new_end_point=(0, edit_offset + 2),
        )

        new_tree = parser.parse(b"def foo(ab):\n  bar()", tree)
        changed_ranges = tree.changed_ranges(new_tree)

        self.assertEqual(len(changed_ranges), 1)
        self.assertEqual(changed_ranges[0].start_byte, edit_offset)
        self.assertEqual(changed_ranges[0].start_point, (0, edit_offset))
        self.assertEqual(changed_ranges[0].end_byte, edit_offset + 2)
        self.assertEqual(changed_ranges[0].end_point, (0, edit_offset + 2))

    def test_walk(self):
        parser = Parser(self.rust)

        tree = parser.parse(
            b"""
                struct Stuff {
                    a: A,
                    b: Option<B>,
                }
            """
        )

        cursor = tree.walk()

        # Node always returns the same instance
        self.assertIs(cursor.node, cursor.node)

        self.assertEqual(cast(Node, cursor.node).type, "source_file")

        self.assertEqual(cursor.goto_first_child(), True)
        self.assertEqual(cast(Node, cursor.node).type, "struct_item")

        self.assertEqual(cursor.goto_first_child(), True)
        self.assertEqual(cast(Node, cursor.node).type, "struct")
        self.assertEqual(cast(Node, cursor.node).is_named, False)

        self.assertEqual(cursor.goto_next_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, "type_identifier")
        self.assertEqual(cast(Node, cursor.node).is_named, True)

        self.assertEqual(cursor.goto_next_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, "field_declaration_list")
        self.assertEqual(cast(Node, cursor.node).is_named, True)

        self.assertEqual(cursor.goto_last_child(), True)
        self.assertEqual(cast(Node, cursor.node).type, "}")
        self.assertEqual(cast(Node, cursor.node).is_named, False)
        self.assertEqual(cast(Node, cursor.node).start_point, (4, 16))

        self.assertEqual(cursor.goto_previous_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, ",")
        self.assertEqual(cast(Node, cursor.node).is_named, False)
        self.assertEqual(cast(Node, cursor.node).start_point, (3, 32))

        self.assertEqual(cursor.goto_previous_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, "field_declaration")
        self.assertEqual(cast(Node, cursor.node).is_named, True)
        self.assertEqual(cast(Node, cursor.node).start_point, (3, 20))

        self.assertEqual(cursor.goto_previous_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, ",")
        self.assertEqual(cast(Node, cursor.node).is_named, False)
        self.assertEqual(cast(Node, cursor.node).start_point, (2, 24))

        self.assertEqual(cursor.goto_previous_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, "field_declaration")
        self.assertEqual(cast(Node, cursor.node).is_named, True)
        self.assertEqual(cast(Node, cursor.node).start_point, (2, 20))

        self.assertEqual(cursor.goto_previous_sibling(), True)
        self.assertEqual(cast(Node, cursor.node).type, "{")
        self.assertEqual(cast(Node, cursor.node).is_named, False)
        self.assertEqual(cast(Node, cursor.node).start_point, (1, 29))

        copy = tree.walk()
        copy.reset_to(cursor)

        self.assertEqual(cast(Node, copy.node).type, "{")
        self.assertEqual(cast(Node, copy.node).is_named, False)

        self.assertEqual(copy.goto_parent(), True)
        self.assertEqual(cast(Node, copy.node).type, "field_declaration_list")
        self.assertEqual(cast(Node, copy.node).is_named, True)

        self.assertEqual(copy.goto_parent(), True)
        self.assertEqual(cast(Node, copy.node).type, "struct_item")



from re import error as RegexError
from unittest import TestCase

import tree_sitter_python
import tree_sitter_javascript

from tree_sitter import Language, Parser, Query, QueryCursor, QueryError


def collect_matches(matches):
    return [(m[0], format_captures(m[1])) for m in matches]


def format_captures(captures):
    return [(name, format_capture(capture)) for name, capture in captures.items()]


def format_capture(capture):
    return [n.text.decode("utf-8") for n in capture]


class TestQuery(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.javascript = Language(tree_sitter_javascript.language())
        cls.python = Language(tree_sitter_python.language())

    def assert_query_matches(self, language, query, source, expected):
        parser = Parser(language)
        tree = parser.parse(source)
        cursor = QueryCursor(Query(language, query))
        matches = cursor.matches(tree.root_node)
        self.assertListEqual(collect_matches(matches), expected)

    def test_errors(self):
        with self.assertRaises(QueryError):
            Query(self.python, "(list (foo))")
        with self.assertRaises(QueryError):
            Query(self.python, "(function_definition buzz: (identifier))")
        with self.assertRaises(QueryError):
            Query(self.python, "((function_definition) (#eq? @garbage foo))")
        with self.assertRaises(QueryError):
            Query(self.python, "(list))")

    def test_matches_with_simple_pattern(self):
        self.assert_query_matches(
            self.javascript,
            "(function_declaration name: (identifier) @fn-name)",
            b"function one() { two(); function three() {} }",
            [(0, [("fn-name", ["one"])]), (0, [("fn-name", ["three"])])],
        )

    def test_matches_with_multiple_on_same_root(self):
        self.assert_query_matches(
            self.javascript,
            """
            (class_declaration
                name: (identifier) @the-class-name
                (class_body
                    (method_definition
                        name: (property_identifier) @the-method-name)))
            """,
            b"""
            class Person {
                // the constructor
                constructor(name) { this.name = name; }

                // the getter
                getFullName() { return this.name; }
            }
            """,
            [
                (0, [("the-class-name", ["Person"]), ("the-method-name", ["constructor"])]),
                (0, [("the-class-name", ["Person"]), ("the-method-name", ["getFullName"])]),
            ],
        )

    def test_matches_with_multiple_patterns_different_roots(self):
        self.assert_query_matches(
            self.javascript,
            """
            (function_declaration name: (identifier) @fn-def)
            (call_expression function: (identifier) @fn-ref)
            """,
            b"""
            function f1() {
                f2(f3());
            }
            """,
            [
                (0, [("fn-def", ["f1"])]),
                (1, [("fn-ref", ["f2"])]),
                (1, [("fn-ref", ["f3"])]),
            ],
        )

    def test_matches_with_nesting_and_no_fields(self):
        self.assert_query_matches(
            self.javascript,
            "(array (array (identifier) @x1 (identifier) @x2))",
            b"""
            [[a]];
            [[c, d], [e, f, g, h]];
            [[h], [i]];
            """,
            [
                (0, [("x1", ["c"]), ("x2", ["d"])]),
                (0, [("x1", ["e"]), ("x2", ["f"])]),
                (0, [("x1", ["e"]), ("x2", ["g"])]),
                (0, [("x1", ["f"]), ("x2", ["g"])]),
                (0, [("x1", ["e"]), ("x2", ["h"])]),
                (0, [("x1", ["f"]), ("x2", ["h"])]),
                (0, [("x1", ["g"]), ("x2", ["h"])]),
            ],
        )

    def test_matches_with_list_capture(self):
        self.assert_query_matches(
            self.javascript,
            """
            (function_declaration
                name: (identifier) @fn-name
                body: (statement_block (_)* @fn-statements))
            """,
            b"""function one() {
                    x = 1;
                    y = 2;
                    z = 3;
                }
                function two() {
                    x = 1;
                }
            """,
            [
                (
                    0,
                    [
                        ("fn-name", ["one"]),
                        ("fn-statements", ["x = 1;", "y = 2;", "z = 3;"]),
                    ],
                ),
                (0, [("fn-name", ["two"]), ("fn-statements", ["x = 1;"])]),
            ],
        )

    def test_captures(self):
        parser = Parser(self.python)
        source = b"def foo():\n  bar()\ndef baz():\n  quux()\n"
        tree = parser.parse(source)
        query = Query(
            self.python,
            """
            (function_definition name: (identifier) @func-def)
            (call function: (identifier) @func-call)
            """,
        )

        cursor = QueryCursor(query)
        captures = list(cursor.captures(tree.root_node).items())

        self.assertEqual(captures[0][0], "func-def")
        self.assertEqual(captures[0][1][0].start_point, (0, 4))
        self.assertEqual(captures[0][1][0].end_point, (0, 7))
        self.assertEqual(captures[0][1][1].start_point, (2, 4))
        self.assertEqual(captures[0][1][1].end_point, (2, 7))

        self.assertEqual(captures[1][0], "func-call")
        self.assertEqual(captures[1][1][0].start_point, (1, 2))
        self.assertEqual(captures[1][1][0].end_point, (1, 5))
        self.assertEqual(captures[1][1][1].start_point, (3, 2))
        self.assertEqual(captures[1][1][1].end_point, (3, 6))

    def test_text_predicates(self):
        parser = Parser(self.javascript)
        source = b"""
            keypair_object = {
                key1: value1,
                equal: equal
            }

            function fun1(arg) {
                return 1;
            }

            function fun2(arg) {
                return 2;
            }
        """
        tree = parser.parse(source)
        root_node = tree.root_node

        # function with name equal to 'fun1' -> test for #eq? @capture string
        query1 = Query(
            self.javascript,
            """
            ((function_declaration
                name: (identifier) @function-name)
                (#eq? @function-name fun1))
            """
        )
        cursor = QueryCursor(query1)
        captures1 = list(cursor.captures(root_node).items())
        self.assertEqual(1, len(captures1))
        self.assertEqual(captures1[0][0], "function-name")
        self.assertEqual(captures1[0][1][0].text, b"fun1")

        # functions with name not equal to 'fun1' -> test for #not-eq? @capture string
        query2 = Query(
            self.javascript,
            """
            ((function_declaration
                name: (identifier) @function-name)
                (#not-eq? @function-name fun1))
            """
        )
        cursor = QueryCursor(query2)
        captures2 = list(cursor.captures(root_node).items())
        self.assertEqual(1, len(captures2))
        self.assertEqual(captures2[0][0], "function-name")
        self.assertEqual(captures2[0][1][0].text, b"fun2")

    def test_text_predicates_with_callback(self):
        parser = Parser(self.javascript)
        source = b"""
            keypair_object = {
                key1: value1,
                equal: equal
            }

            function fun1(arg) {
                return 1;
            }

            function fun2(arg) {
                return 2;
            }

            function fun3(arg) {
                return 3;
            }
        """

        def read_callable_byte_offset(byte_offset, point):
            return source[byte_offset: byte_offset + 1]

        def read_callable_point(byte_offset, point):
            row, col = point
            lines = source.split(b"\n")
            if row >= len(lines):
                return b""
            line = lines[row]
            if col >= len(line):
                return b"\n"
            return line[col:col + 1]

        tree1 = parser.parse(read_callable_byte_offset)
        root_node1 = tree1.root_node
        tree2 = parser.parse(read_callable_point)
        root_node2 = tree2.root_node

        # function with name equal to 'fun1' -> test for #eq? @capture string
        query1 = Query(
            self.javascript,
            """
            ((function_declaration
                name: (identifier) @function-name)
                (#match? @function-name "fun[12]"))
            """
        )
        cursor1 = QueryCursor(query1)
        captures1 = cursor1.captures(root_node1)
        self.assertEqual(1, len(captures1))
        self.assertIn("function-name", captures1)
        self.assertEqual(2, len(captures1["function-name"]))
        self.assertEqual(captures1["function-name"][0].text, b"fun1")
        self.assertEqual(captures1["function-name"][1].text, b"fun2")

        captures2 = cursor1.captures(root_node2)
        self.assertEqual(1, len(captures2))
        self.assertIn("function-name", captures2)
        self.assertEqual(2, len(captures2["function-name"]))
        self.assertEqual(captures2["function-name"][0].text, b"fun1")
        self.assertEqual(captures2["function-name"][1].text, b"fun2")

        # functions with name not equal to 'fun1' -> test for #not-eq? @capture string
        query2 = Query(
            self.javascript,
            """
            ((function_declaration
                name: (identifier) @function-name)
                (#not-eq? @function-name fun1))
            """
        )
        cursor2 = QueryCursor(query2)
        captures3 = cursor2.captures(root_node1)
        self.assertEqual(1, len(captures3))
        self.assertIn("function-name", captures3)
        self.assertEqual(2, len(captures3["function-name"]))
        self.assertEqual(captures3["function-name"][0].text, b"fun2")
        self.assertEqual(captures3["function-name"][1].text, b"fun3")

        captures4 = cursor2.captures(root_node2)
        self.assertEqual(1, len(captures4))
        self.assertIn("function-name", captures4)
        self.assertEqual(2, len(captures4["function-name"]))
        self.assertEqual(captures4["function-name"][0].text, b"fun2")
        self.assertEqual(captures4["function-name"][1].text, b"fun3")

    def test_text_predicates_errors(self):
        with self.assertRaises(QueryError):
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#eq? @function-name @function-name fun1))
                """
            )

        with self.assertRaises(QueryError):
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#eq? fun1 @function-name))
                """
            )

        with self.assertRaises(QueryError):
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#match? @function-name @function-name fun1))
                """
            )

        with self.assertRaises(QueryError):
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#match? fun1 @function-name))
                """
            )

        with self.assertRaises(QueryError):
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#match? @function-name @function-name))
                """
            )

        with self.assertRaises(QueryError) as ctx:
            Query(
                self.javascript,
                """
                ((function_declaration
                    name: (identifier) @function-name)
                    (#match? @function-name "?"))
                """
            )
        self.assertEqual(
            str(ctx.exception), "Invalid predicate in pattern at row 1: regular expression error"
        )
        self.assertIsInstance(ctx.exception.__cause__, RegexError)

    def test_point_range_captures(self):
        parser = Parser(self.python)
        source = b"def foo():\n  bar()\ndef baz():\n  quux()\n"
        tree = parser.parse(source)
        query = Query(
            self.python,
            """
            (function_definition name: (identifier) @func-def)
            (call function: (identifier) @func-call)
            """
        )
        cursor = QueryCursor(query)
        cursor.set_point_range((1, 0), (2, 0))

        captures = list(cursor.captures(tree.root_node).items())

        self.assertEqual(captures[0][0], "func-call")
        self.assertEqual(captures[0][1][0].start_point, (1, 2))
        self.assertEqual(captures[0][1][0].end_point, (1, 5))

    def test_byte_range_captures(self):
        parser = Parser(self.python)
        source = b"def foo():\n  bar()\ndef baz():\n  quux()\n"
        tree = parser.parse(source)
        query = Query(
            self.python,
            """
            (function_definition name: (identifier) @func-def)
            (call function: (identifier) @func-call)
            """
        )
        cursor = QueryCursor(query)
        cursor.set_byte_range(10, 20)

        captures = list(cursor.captures(tree.root_node).items())
        self.assertEqual(captures[0][0], "func-call")
        self.assertEqual(captures[0][1][0].start_point, (1, 2))
        self.assertEqual(captures[0][1][0].end_point, (1, 5))



from typing import cast
from unittest import TestCase

from tree_sitter import Language, LookaheadIterator, Node, Parser

import tree_sitter_rust


class TestLookaheadIterator(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rust = Language(tree_sitter_rust.language())

    def test_lookahead_iterator(self):
        parser = Parser(self.rust)
        cursor = parser.parse(b"struct Stuff{}").walk()

        self.assertEqual(cursor.goto_first_child(), True)  # struct
        self.assertEqual(cursor.goto_first_child(), True)  # struct keyword

        node = cast(Node, cursor.node)
        next_state = node.next_parse_state

        self.assertNotEqual(next_state, 0)
        self.assertEqual(
            next_state, self.rust.next_state(node.parse_state, node.grammar_id)
        )
        self.assertLess(next_state, self.rust.parse_state_count)
        self.assertEqual(cursor.goto_next_sibling(), True)  # type_identifier
        node = cast(Node, cursor.node)
        self.assertEqual(next_state, node.parse_state)
        self.assertEqual(node.grammar_name, "identifier")
        self.assertNotEqual(node.grammar_id, node.kind_id)

        expected_symbols = ["//", "/*", "identifier", "line_comment", "block_comment"]
        lookahead = cast(LookaheadIterator, self.rust.lookahead_iterator(next_state))
        self.assertEqual(lookahead.language, self.rust)
        self.assertListEqual(lookahead.names(), expected_symbols)

        lookahead.reset(next_state)
        self.assertListEqual(
            list(map(self.rust.node_kind_for_id, lookahead.symbols())), expected_symbols
        )

        lookahead.reset(next_state, self.rust)
        self.assertTupleEqual(
            (self.rust.id_for_node_kind("//", False), "//"),
            next(lookahead)
        )



from typing import cast
from unittest import TestCase

from tree_sitter import Language, LogType, Node, Parser, Range, Tree

import tree_sitter_html
import tree_sitter_javascript
import tree_sitter_json
import tree_sitter_python
import tree_sitter_rust


def simple_range(start, end):
    return Range((0, start), (0, end), start, end)


class TestParser(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.html = Language(tree_sitter_html.language())
        cls.python = Language(tree_sitter_python.language())
        cls.javascript = Language(tree_sitter_javascript.language())
        cls.json = Language(tree_sitter_json.language())
        cls.rust = Language(tree_sitter_rust.language())
        cls.max_range = Range((0, 0), (0xFFFFFFFF, 0xFFFFFFFF), 0, 0xFFFFFFFF)
        cls.min_range = Range((0, 0), (0, 1), 0, 1)
        cls.timeout = 1000

    def test_init_no_args(self):
        parser = Parser()
        self.assertIsNone(parser.language)
        self.assertListEqual(parser.included_ranges, [self.max_range])

    def test_init_args(self):
        parser = Parser(
            language=self.python, included_ranges=[self.min_range]
        )
        self.assertEqual(parser.language, self.python)
        self.assertListEqual(parser.included_ranges, [self.min_range])

    def test_setters(self):
        parser = Parser()

        with self.subTest(setter="language"):
            parser.language = self.python
            self.assertEqual(parser.language, self.python)

        with self.subTest(setter="included_ranges"):
            parser.included_ranges = [self.min_range]
            self.assertListEqual(parser.included_ranges, [self.min_range])
            with self.assertRaises(ValueError):
                parser.included_ranges = [
                    Range(
                        start_byte=23,
                        end_byte=29,
                        start_point=(0, 23),
                        end_point=(0, 29),
                    ),
                    Range(
                        start_byte=0,
                        end_byte=5,
                        start_point=(0, 0),
                        end_point=(0, 5),
                    ),
                    Range(
                        start_byte=50,
                        end_byte=60,
                        start_point=(0, 50),
                        end_point=(0, 60),
                    ),
                ]
            with self.assertRaises(ValueError):
                parser.included_ranges = [
                    Range(
                        start_byte=10,
                        end_byte=5,
                        start_point=(0, 10),
                        end_point=(0, 5),
                    )
                ]

        with self.subTest(setter="logger"):
            def logger(log_type, message):
                print(log_type.name, message)

            parser.logger = logger
            self.assertEqual(parser.logger, logger)

    def test_deleters(self):
        parser = Parser()

        with self.subTest(deleter="language"):
            del parser.language
            self.assertIsNone(parser.language)

        with self.subTest(deleter="included_ranges"):
            del parser.included_ranges
            self.assertListEqual(parser.included_ranges, [self.max_range])

        with self.subTest(deleter="logger"):
            del parser.logger
            self.assertEqual(parser.logger, None)

    def test_parse_buffer(self):
        parser = Parser(self.javascript)
        with self.subTest(type="bytes"):
            self.assertIsInstance(parser.parse(b"test"), Tree)
        with self.subTest(type="memoryview"):
            self.assertIsInstance(parser.parse(memoryview(b"test")), Tree)
        with self.subTest(type="bytearray"):
            self.assertIsInstance(parser.parse(bytearray(b"test")), Tree)

    def test_parse_callback(self):
        parser = Parser(self.python)
        source_lines = ["def foo():\n", "  bar()"]

        def read_callback(_, point):
            row, column = point
            if row >= len(source_lines):
                return None
            if column >= len(source_lines[row]):
                return None
            return source_lines[row][column:].encode("utf8")

        def modify_return(cb, ret_type):
            def wrapped(*args):
                rv = cb(*args)
                return ret_type(rv) if rv is not None else None
            return wrapped

        def assert_nodes_equal(tree: Tree):
            self.assertEqual(
                str(tree.root_node),
                "(module (function_definition"
                + " name: (identifier)"
                + " parameters: (parameters)"
                + " body: (block (expression_statement (call"
                + " function: (identifier)"
                + " arguments: (argument_list))))))",
            )

        with self.subTest(type="bytes"):
            tree = parser.parse(read_callback)
            assert_nodes_equal(tree)
        with self.subTest(type="memoryview"):
            tree = parser.parse(modify_return(read_callback, memoryview))
            assert_nodes_equal(tree)
        with self.subTest(type="bytearray"):
            tree = parser.parse(modify_return(read_callback, bytearray))
            assert_nodes_equal(tree)

    def test_parse_utf16_encoding(self):
        source_code = bytes("'😎' && '🐍'", "utf16")
        parser = Parser(self.javascript)

        def read(byte_position, _):
            return source_code[byte_position : byte_position + 2]

        tree = parser.parse(read, encoding="utf16")
        root_node = tree.root_node
        snake_node = root_node.children[0].children[0].children[2]
        snake = source_code[snake_node.start_byte + 2 : snake_node.end_byte - 2]

        self.assertEqual(snake_node.type, "string")
        self.assertEqual(snake.decode("utf16"), "🐍")
        self.assertIs(tree.language, self.javascript)

    def test_parse_invalid_encoding(self):
        parser = Parser(self.python)
        with self.assertRaises(ValueError):
            parser.parse(b"foo", encoding="ascii")  # pyright: ignore

    def test_parse_with_one_included_range(self):
        source_code = b"<span>hi</span><script>console.log('sup');</script>"
        parser = Parser(self.html)
        html_tree = parser.parse(source_code)
        script_content_node = cast(Node, html_tree.root_node.child(1)).child(1)
        self.assertIsNotNone(script_content_node)
        script_content_node = cast(Node, script_content_node)
        self.assertEqual(script_content_node.type, "raw_text")

        parser.included_ranges = [script_content_node.range]
        parser.language = self.javascript
        js_tree = parser.parse(source_code)
        self.assertEqual(
            str(js_tree.root_node),
            "(program (expression_statement (call_expression"
            + " function: (member_expression object: (identifier) property: (property_identifier))"
            + " arguments: (arguments (string (string_fragment))))))",
        )
        self.assertEqual(js_tree.root_node.start_point, (0, source_code.index(b"console")))
        self.assertEqual(js_tree.included_ranges, [script_content_node.range])

    def test_parse_with_multiple_included_ranges(self):
        source_code = b"html `<div>Hello, ${name.toUpperCase()}, it's <b>${now()}</b>.</div>`"

        parser = Parser(self.javascript)
        js_tree = parser.parse(source_code)
        template_string_node = js_tree.root_node.descendant_for_byte_range(
            source_code.index(b"`<"), source_code.index(b">`")
        )
        self.assertIsNotNone(template_string_node)
        template_string_node = cast(Node, template_string_node)

        self.assertEqual(template_string_node.type, "template_string")

        open_quote_node = cast(Node, template_string_node.child(0))
        self.assertIsNotNone(open_quote_node)
        interpolation_node1 = cast(Node, template_string_node.child(2))
        self.assertIsNotNone(interpolation_node1)
        interpolation_node2 = cast(Node, template_string_node.child(4))
        self.assertIsNotNone(interpolation_node2)
        close_quote_node = cast(Node, template_string_node.child(6))
        self.assertIsNotNone(close_quote_node)

        html_ranges = [
            Range(
                start_byte=open_quote_node.end_byte,
                start_point=open_quote_node.end_point,
                end_byte=interpolation_node1.start_byte,
                end_point=interpolation_node1.start_point,
            ),
            Range(
                start_byte=interpolation_node1.end_byte,
                start_point=interpolation_node1.end_point,
                end_byte=interpolation_node2.start_byte,
                end_point=interpolation_node2.start_point,
            ),
            Range(
                start_byte=interpolation_node2.end_byte,
                start_point=interpolation_node2.end_point,
                end_byte=close_quote_node.start_byte,
                end_point=close_quote_node.start_point,
            ),
        ]
        parser.included_ranges = html_ranges
        parser.language = self.html
        html_tree = parser.parse(source_code)

        self.assertEqual(
            str(html_tree.root_node),
            "(document (element"
            + " (start_tag (tag_name))"
            + " (text)"
            + " (element (start_tag (tag_name)) (end_tag (tag_name)))"
            + " (text)"
            + " (end_tag (tag_name))))",
        )
        self.assertEqual(html_tree.included_ranges, html_ranges)

        div_element_node = cast(Node, html_tree.root_node.child(0))
        self.assertIsNotNone(div_element_node)
        hello_text_node = cast(Node, div_element_node.child(1))
        self.assertIsNotNone(hello_text_node)
        b_element_node = cast(Node, div_element_node.child(2))
        self.assertIsNotNone(b_element_node)
        b_start_tag_node = cast(Node, b_element_node.child(0))
        self.assertIsNotNone(b_start_tag_node)
        b_end_tag_node = cast(Node, b_element_node.child(1))
        self.assertIsNotNone(b_end_tag_node)

        self.assertEqual(hello_text_node.type, "text")
        self.assertEqual(hello_text_node.start_byte, source_code.index(b"Hello"))
        self.assertEqual(hello_text_node.end_byte, source_code.index(b" <b>"))

        self.assertEqual(b_start_tag_node.type, "start_tag")
        self.assertEqual(b_start_tag_node.start_byte, source_code.index(b"<b>"))
        self.assertEqual(b_start_tag_node.end_byte, source_code.index(b"${now()}"))

        self.assertEqual(b_end_tag_node.type, "end_tag")
        self.assertEqual(b_end_tag_node.start_byte, source_code.index(b"</b>"))
        self.assertEqual(b_end_tag_node.end_byte, source_code.index(b".</div>"))

    def test_parse_with_included_range_containing_mismatched_positions(self):
        source_code = b"<div>test</div>{_ignore_this_part_}"
        end_byte = source_code.index(b"{_ignore_this_part_")

        range_to_parse = Range(
            start_byte=0,
            start_point=(10, 12),
            end_byte=end_byte,
            end_point=(10, 12 + end_byte),
        )

        parser = Parser(self.html, included_ranges=[range_to_parse])
        html_tree = parser.parse(source_code)

        self.assertEqual(
            str(html_tree.root_node),
            "(document (element (start_tag (tag_name)) (text) (end_tag (tag_name))))",
        )

    def test_parse_with_included_range_boundaries(self):
        source_code = b"a <%= b() %> c <% d() %>"
        range1_start_byte = source_code.index(b" b() ")
        range1_end_byte = range1_start_byte + len(b" b() ")
        range2_start_byte = source_code.index(b" d() ")
        range2_end_byte = range2_start_byte + len(b" d() ")

        parser = Parser(
            self.javascript,
            included_ranges=[
                Range(
                    start_byte=range1_start_byte,
                    end_byte=range1_end_byte,
                    start_point=(0, range1_start_byte),
                    end_point=(0, range1_end_byte),
                ),
                Range(
                    start_byte=range2_start_byte,
                    end_byte=range2_end_byte,
                    start_point=(0, range2_start_byte),
                    end_point=(0, range2_end_byte),
                ),
            ],
        )

        tree = parser.parse(source_code)
        root = tree.root_node
        statement1 = cast(Node, root.child(0))
        self.assertIsNotNone(statement1)
        statement2 = cast(Node, root.child(1))
        self.assertIsNotNone(statement2)

        self.assertEqual(
            str(root),
            "(program"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments)))"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments))))",
        )

        self.assertEqual(statement1.start_byte, source_code.index(b"b()"))
        self.assertEqual(statement1.end_byte, source_code.find(b" %> c"))
        self.assertEqual(statement2.start_byte, source_code.find(b"d()"))
        self.assertEqual(statement2.end_byte, len(source_code) - len(" %>"))

    def test_parse_with_a_newly_excluded_range(self):
        source_code = b"<div><span><%= something %></span></div>"

        # Parse HTML including the template directive, which will cause an error
        parser = Parser(self.html)
        first_tree = parser.parse(source_code)

        prefix = b"a very very long line of plain text. "
        first_tree.edit(
            start_byte=0,
            old_end_byte=0,
            new_end_byte=len(prefix),
            start_point=(0, 0),
            old_end_point=(0, 0),
            new_end_point=(0, len(prefix)),
        )
        source_code = prefix + source_code

        # Parse the HTML again, this time *excluding* the template directive
        # (which has moved since the previous parse).
        directive_start = source_code.index(b"<%=")
        directive_end = source_code.index(b"</span>")
        source_code_end = len(source_code)
        parser.included_ranges = [
            Range(
                start_byte=0,
                end_byte=directive_start,
                start_point=(0, 0),
                end_point=(0, directive_start),
            ),
            Range(
                start_byte=directive_end,
                end_byte=source_code_end,
                start_point=(0, directive_end),
                end_point=(0, source_code_end),
            ),
        ]

        tree = parser.parse(source_code, first_tree)

        self.assertEqual(
            str(tree.root_node),
            "(document (text) (element"
            + " (start_tag (tag_name))"
            + " (element (start_tag (tag_name)) (end_tag (tag_name)))"
            + " (end_tag (tag_name))))",
        )

        self.assertEqual(
            tree.changed_ranges(first_tree),
            [
                # The first range that has changed syntax is the range of the newly-inserted text.
                Range(
                    start_byte=0,
                    end_byte=len(prefix),
                    start_point=(0, 0),
                    end_point=(0, len(prefix)),
                ),
                # Even though no edits were applied to the outer `div` element,
                # its contents have changed syntax because a range of text that
                # was previously included is now excluded.
                Range(
                    start_byte=directive_start,
                    end_byte=directive_end,
                    start_point=(0, directive_start),
                    end_point=(0, directive_end),
                ),
            ],
        )

    def test_parsing_with_a_newly_included_range(self):
        source_code = b"<div><%= foo() %></div><span><%= bar() %></span><%= baz() %>"
        range1_start = source_code.index(b" foo")
        range2_start = source_code.index(b" bar")
        range3_start = source_code.index(b" baz")
        range1_end = range1_start + 7
        range2_end = range2_start + 7
        range3_end = range3_start + 7

        # Parse only the first code directive as JavaScript
        parser = Parser(self.javascript)
        parser.included_ranges = [simple_range(range1_start, range1_end)]
        tree = parser.parse(source_code)
        self.assertEqual(
            str(tree.root_node),
            "(program"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments))))",
        )

        # Parse both the first and third code directives as JavaScript, using the old tree as a
        # reference.
        parser.included_ranges = [
            simple_range(range1_start, range1_end),
            simple_range(range3_start, range3_end),
        ]
        tree2 = parser.parse(source_code)
        self.assertEqual(
            str(tree2.root_node),
            "(program"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments)))"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments))))",
        )
        self.assertEqual(tree2.changed_ranges(tree), [simple_range(range1_end, range3_end)])

        # Parse all three code directives as JavaScript, using the old tree as a
        # reference.
        parser.included_ranges = [
            simple_range(range1_start, range1_end),
            simple_range(range2_start, range2_end),
            simple_range(range3_start, range3_end),
        ]
        tree3 = parser.parse(source_code)
        self.assertEqual(
            str(tree3.root_node),
            "(program"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments)))"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments)))"
            + " (expression_statement (call_expression"
            + " function: (identifier) arguments: (arguments))))",
        )
        self.assertEqual(
            tree3.changed_ranges(tree2),
            [simple_range(range2_start + 1, range2_end - 1)],
        )

    def test_logging(self):
        from logging import getLogger

        def logger(log_type: LogType, message: str):
            match log_type:
                case LogType.PARSE:
                    parse_logger.info(message)
                case LogType.LEX:
                    lex_logger.info(message)

        parse_logger = getLogger("tree_sitter.PARSE")
        lex_logger = getLogger("tree_sitter.LEX")
        parser = Parser(self.python, logger=logger)
        with self.assertLogs("tree_sitter") as logs:
            parser.parse(b"foo")

        self.assertEqual(logs.records[0].name, "tree_sitter.PARSE")
        self.assertEqual(logs.records[0].message, "new_parse")
        self.assertEqual(logs.records[3].name, "tree_sitter.LEX")
        self.assertEqual(logs.records[3].message, "consume character:'f'")

    def test_dot_graphs(self):
        from tempfile import TemporaryFile

        new_parse = ["graph {\n", 'label="new_parse"\n', "}\n"]
        parser = Parser(self.python)
        with TemporaryFile("w+") as f:
            parser.print_dot_graphs(f)
            parser.parse(b"foo")
            f.seek(0)
            lines = [f.readline(), f.readline(), f.readline()]
            self.assertListEqual(lines, new_parse)



from tree_sitter import Language, Parser, Query, QueryCursor
import tree_sitter_python

PY_LANGUAGE = Language(tree_sitter_python.language())

parser = Parser(PY_LANGUAGE)

# parsing a string of code
tree = parser.parse(
    bytes(
        """
def foo():
    if bar:
        baz()
""",
        "utf8",
    )
)

# parsing a callable by using the byte offset
src = bytes(
    """
def foo():
    if bar:
        baz()
""",
    "utf8",
)


def read_callable_byte_offset(byte_offset, _):
    return src[byte_offset : byte_offset + 1]


tree = parser.parse(read_callable_byte_offset)


# parsing a callable by using the point
src_lines = ["\n", "def foo():\n", "    if bar:\n", "        baz()\n"]


def read_callable_point(_, point):
    row, column = point
    if row >= len(src_lines) or column >= len(src_lines[row]):
        return None
    return src_lines[row][column:].encode("utf8")


tree = parser.parse(read_callable_point)

# inspecting nodes in the tree
root_node = tree.root_node
assert root_node.type == "module"
assert root_node.start_point == (1, 0)
assert root_node.end_point == (4, 0)

function_node = root_node.child(0)
assert function_node.type == "function_definition"
assert function_node.child_by_field_name("name").type == "identifier"

function_name_node = function_node.child(1)
assert function_name_node.type == "identifier"
assert function_name_node.start_point == (1, 4)
assert function_name_node.end_point == (1, 7)

function_body_node = function_node.child_by_field_name("body")

if_statement_node = function_body_node.child(0)
assert if_statement_node.type == "if_statement"

function_call_node = if_statement_node.child_by_field_name("consequence").child(0).child(0)
assert function_call_node.type == "call"

function_call_name_node = function_call_node.child_by_field_name("function")
assert function_call_name_node.type == "identifier"

function_call_args_node = function_call_node.child_by_field_name("arguments")
assert function_call_args_node.type == "argument_list"


# getting the sexp representation of the tree
assert str(root_node) == (
    "(module "
    "(function_definition "
    "name: (identifier) "
    "parameters: (parameters) "
    "body: (block "
    "(if_statement "
    "condition: (identifier) "
    "consequence: (block "
    "(expression_statement (call "
    "function: (identifier) "
    "arguments: (argument_list))))))))"
)

# walking the tree
cursor = tree.walk()

assert cursor.node.type == "module"

assert cursor.goto_first_child()
assert cursor.node.type == "function_definition"

assert cursor.goto_first_child()
assert cursor.node.type == "def"

# Returns `False` because the `def` node has no children
assert not cursor.goto_first_child()

assert cursor.goto_next_sibling()
assert cursor.node.type == "identifier"

assert cursor.goto_next_sibling()
assert cursor.node.type == "parameters"

assert cursor.goto_parent()
assert cursor.node.type == "function_definition"

# editing the tree
new_src = src[:5] + src[5 : 5 + 2].upper() + src[5 + 2 :]

tree.edit(
    start_byte=5,
    old_end_byte=5,
    new_end_byte=5 + 2,
    start_point=(0, 5),
    old_end_point=(0, 5),
    new_end_point=(0, 5 + 2),
)

new_tree = parser.parse(new_src, tree)

# inspecting the changes
for changed_range in tree.changed_ranges(new_tree):
    print("Changed range:")
    print(f"  Start point {changed_range.start_point}")
    print(f"  Start byte {changed_range.start_byte}")
    print(f"  End point {changed_range.end_point}")
    print(f"  End byte {changed_range.end_byte}")


# querying the tree
query = Query(
    PY_LANGUAGE,
    """
(function_definition
  name: (identifier) @function.def
  body: (block) @function.block)

(call
  function: (identifier) @function.call
  arguments: (argument_list) @function.args)
""",
)

# ...with captures
query_cursor = QueryCursor(query)
captures = query_cursor.captures(tree.root_node)
assert len(captures) == 4
assert captures["function.def"][0] == function_name_node
assert captures["function.block"][0] == function_body_node
assert captures["function.call"][0] == function_call_name_node
assert captures["function.args"][0] == function_call_args_node

# ...with matches
matches = query_cursor.matches(tree.root_node)
assert len(matches) == 2

# first match
assert matches[0][1]["function.def"] == [function_name_node]
assert matches[0][1]["function.block"] == [function_body_node]

# second match
assert matches[1][1]["function.call"] == [function_call_name_node]
assert matches[1][1]["function.args"] == [function_call_args_node]



from typing import Generator
from tree_sitter import Language, Parser, Tree, Node
import tree_sitter_python

PY_LANGUAGE = Language(tree_sitter_python.language())

parser = Parser()
parser.language = PY_LANGUAGE

tree = parser.parse(bytes("a = 1", "utf8"))


def traverse_tree(tree: Tree) -> Generator[Node, None, None]:
    cursor = tree.walk()

    visited_children = False
    while True:
        if not visited_children:
            yield cursor.node
            if not cursor.goto_first_child():
                visited_children = True
        elif cursor.goto_next_sibling():
            visited_children = False
        elif not cursor.goto_parent():
            break


node_names = map(lambda node: node.type, traverse_tree(tree))

assert list(node_names) == [
    "module",
    "expression_statement",
    "assignment",
    "identifier",
    "=",
    "integer",
]
